package org.ow2.orchestra.osgi;

import java.lang.management.ManagementFactory;
import java.net.URL;
import java.rmi.server.ExportException;
import java.util.Dictionary;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;
import java.util.logging.Level;

import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXConnectorServerFactory;
import javax.management.remote.JMXServiceURL;
import javax.servlet.Servlet;
import javax.servlet.http.HttpServlet;

import org.apache.axis.transport.http.AxisServlet;
import org.osgi.service.http.HttpContext;
import org.osgi.service.http.HttpService;
import org.ow2.orchestra.axis.AxisConfiguration;
import org.ow2.orchestra.env.BpelEnvironmentParser;
import org.ow2.orchestra.env.EnvFactoryRepository;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.jmx.RemoteDeployer;
import org.ow2.orchestra.facade.jmx.RemoteDeployerMBean;
import org.ow2.orchestra.jmx.JMXConfiguration;
import org.ow2.orchestra.jmx.RemoteDeployerImpl;
import org.ow2.orchestra.pvm.env.EnvironmentFactory;
import org.ow2.orchestra.pvm.internal.jobexecutor.JobExecutor;
import org.ow2.orchestra.util.Misc;



public class Engine {

	private final Map<String, Servlet> aliases;

	// Injected by iPOJO
	private HttpService m_httpService;
	private String m_environmentLocation;
	private String m_axisConfigurationLocation;
	private String m_jmxConfigurationLocation;

  private MBeanServer mbs;

  private JMXConnectorServer cs;

  private ObjectName objectName;

	public Engine() {
		this.aliases = new HashMap<String, Servlet>();
	}

	public void setEnvironment(final String environmentLocation) {
		this.m_environmentLocation = environmentLocation;
	}

	public String getEnvironment() {
		return this.m_environmentLocation;
	}


  public String getM_axisConfigurationLocation() {
    return this.m_axisConfigurationLocation;
  }


  public void setM_axisConfigurationLocation(final String configurationLocation) {
    this.m_axisConfigurationLocation = configurationLocation;
  }


  public String getM_jmxConfigurationLocation() {
    return this.m_jmxConfigurationLocation;
  }


  public void setM_jmxConfigurationLocation(final String configurationLocation) {
    this.m_jmxConfigurationLocation = configurationLocation;
  }

	public void start() throws Throwable {
    final ClassLoader bundleClassLoader = this.getClass().getClassLoader();
    final ClassLoader originalClassLoader = Thread.currentThread().getContextClassLoader();
    Thread.currentThread().setContextClassLoader(bundleClassLoader);
		try {
			Misc.log(Level.FINEST, "%s starting", this.getClass().getName());

			// update jmx and axis configurations
			AxisConfiguration.setAxisConfigFile(this.m_axisConfigurationLocation);
			JMXConfiguration.setJmxConfigFile(this.m_jmxConfigurationLocation);

  		// register axisServlet
			final HttpContext context = this.m_httpService.createDefaultHttpContext();
      Misc.log(Level.FINEST, "Context created: " + context);
      this.m_httpService.registerResources("/", "index.html", context);
      final HttpServlet servlet = new AxisServlet();
      Misc.log(Level.FINEST, "Servlet created: %s", servlet);
      final Dictionary<String, String> initparams = new Hashtable<String, String>();
      initparams.put(AxisServlet.INIT_PROPERTY_SERVICES_PATH, "/orchestra/services/");

      this.aliases.put("/orchestra/services", servlet);
      for (final Map.Entry<String, Servlet> mapping : this.aliases.entrySet()) {
        this.m_httpService.registerServlet(mapping.getKey(), mapping.getValue(), initparams , context);
        Misc.log(Level.INFO, "Alias registered: %s for servlet: %s", mapping.getKey(), mapping.getValue());
      }

      // open environment
			if (this.m_environmentLocation == null) {
				Misc.log(Level.WARNING, "Can't get the environmentLocation from OSGi "
						+ "config, using internal environment. This may not be what you are expecting."
						+ " If this is the case, please read your OSGi implementation documentation "
						+ "about ConfigAdmin service.");

				Misc.unreachableStatement("Implement the default behaviour which take the environment from the bundle.");
			}
			Misc.log(Level.INFO, "Environment will be taken from: %s", this.m_environmentLocation);
			final URL url = new URL(this.m_environmentLocation);
			final EnvironmentFactory envFactory = BpelEnvironmentParser.parseEnvironmentFactoryFromURL(url);
			EnvFactoryRepository.set(envFactory);

			// redeploy web services
			new RemoteDeployerImpl().initialize();

			// start job executor
	    final JobExecutor jobExecutor = envFactory.get(JobExecutor.class);
	    if (jobExecutor != null) {
	      jobExecutor.start();
	    }

			// create JMX connexion
	    try {
	      // start rmiregistry if not started
	      try {
	        java.rmi.registry.LocateRegistry.createRegistry(JMXConfiguration.getInstance().getPort());
	      } catch (final ExportException e) {
	        // registry already started, ok
	      }
	      this.mbs = ManagementFactory.getPlatformMBeanServer();
	      final RemoteDeployerMBean remoteDeployer = new RemoteDeployer();
	      this.objectName = ObjectName.getInstance(JMXConfiguration.getInstance().getObjectName());
	      this.mbs.registerMBean(remoteDeployer, this.objectName);
	      final JMXServiceURL jmxUrl = new JMXServiceURL(JMXConfiguration.getInstance().getServiceUrl());
	      this.cs = JMXConnectorServerFactory.newJMXConnectorServer(jmxUrl, null, this.mbs);
	      this.cs.start();

	    } catch (final Exception e) {
	      throw new OrchestraRuntimeException("Problem registering mbean in mbean server",
	          e);
	    }
		} catch (final Throwable t) {
			t.printStackTrace();
			throw t;
		} finally {
	    Thread.currentThread().setContextClassLoader(originalClassLoader);
		}
	}

	public void stop() throws Throwable {
    final ClassLoader bundleClassLoader = this.getClass().getClassLoader();
    final ClassLoader originalClassLoader = Thread.currentThread().getContextClassLoader();
    Thread.currentThread().setContextClassLoader(bundleClassLoader);
		try {
			Misc.log(Level.FINEST, "%s stopping", this.getClass().getName());

			// unregister axis
			for (final String alias : this.aliases.keySet()) {
				this.m_httpService.unregister(alias);
				Misc.log(Level.INFO, "Alias unregistered: %s", alias);
			}
			this.m_httpService.unregister("/");
	     // unregister JMX object
      if (this.mbs != null) {
        try {
          this.cs.stop();
          this.mbs.unregisterMBean(this.objectName);
        } catch (final Exception e) {
          throw new OrchestraRuntimeException("Problem unregistering mbean", e);
        }
      }

			if (EnvFactoryRepository.get() != null) {
			  // stop job executor
	      final JobExecutor jobExecutor = EnvFactoryRepository.get().get(JobExecutor.class);
	      if (jobExecutor != null) {
	        jobExecutor.stop();
	      }
	      // close environment factory
	      EnvFactoryRepository.get().close();
	      EnvFactoryRepository.set(null);
	    }
		} catch (final Throwable t) {
			t.printStackTrace();
			throw t;
		} finally {
      Thread.currentThread().setContextClassLoader(originalClassLoader);
    }
	}
}
